 -------------------------- MODULE Actions ------------------------------------
(* All possible actions of the system. *)

EXTENDS InitialStates

FindNext(i, j) ==
(* Helper function for routing. *)
  IF j = i THEN i
  ELSE
    LET lscan == (LeafSetContent(LeafSets[i]) \ Failed[i]) \ {i}
        next == AbsoluteClosest(j, lscan)
        r == SharedDigits(j, i)
        rtcan == IF r = 0 THEN NullNode
                 \* NA: Condition added for debugging.
                 ELSE RoutingTables[i][<<r, IthDigit(j, r (*+ 1*))>>] \* NA: Should it be IthDigit(j, r)?
        canrelax == {can \in (LeafSetContent(LeafSets[i]) 
                              \cup RTContent(RoutingTables[i]))
                              \ Failed[i]:
                       /\ StrictlyAbsoluteCloser(j, can, i)
                       /\ SharedDigits(j, can) >= r}
    IN 
    IF \/ Overlaps(LeafSets[i])      
       \/ ClockwiseArc(LeftmostNode(LeafSets[i]), j,
                             RightmostNode(LeafSets[i]))
    THEN IF lscan # {}
         THEN IF AbsoluteCloser(j, next, i) THEN next ELSE i
         ELSE i
    ELSE IF rtcan # NullNode
         THEN rtcan
         ELSE IF canrelax # {} THEN AbsoluteClosest(j, canrelax) ELSE i

ProbeSet(i, ls, f, toprobe) == 
(* A set of probe messages to all nodes in toprobe, containing i's leaf set ls
   and the set of failed nodes f. *)
  {[destination |-> j,
    content |-> [type |-> "Probe",  
                 node |-> i,
                 ls |-> ls,
                 failed |-> f]]: j \in toprobe}

-------------------------------------------------------------------------------

(* Consistent Join *)
  
Join(j, i) == 
(* Node j wants to join the network through node i, a "seed" node that
   j knows about from the user application. Node i is the seed node
   responsible for handling join requests from keys numerically closest
   to it. *)
  LET msg == [destination |-> i, 
              content |-> [type |-> "JoinRequest", 
                           rt |-> InitialRT, 
                           node |-> j]] 
  IN
  /\ Status[j] = "Dead"
  /\ Status[i] = "Ready"
  /\ Status' = [Status EXCEPT ![j] = "Waiting"]
  /\ MessagePool' = MessagePool \cup {msg}
  /\ UNCHANGED <<RoutingTables, LeafSets, Probing, Failed, ToJoin, Responsible>>

RouteJoinRequest(i, j) == 
(* Node i forwards a join request addressed to j to a closer node along the
   path. *)
  /\ Status[i] # "Dead"
  /\ \E m \in MessagePool:
       LET nh == FindNext(i, j)
           msg1 == [destination |-> nh,
                    content |-> [type |-> "JoinRequest",
                                 rt |-> AddToRTWithSelf(
                                   RTContent(RoutingTables[i]), m.content.rt, i),
                                 node |-> j]]
           msg2 == [destination |-> i, 
                    content |-> [type |-> "NoLegalRoute", key |-> j]] 
       IN 
       /\ m.content.type = "JoinRequest"
       /\ m.destination = i
       /\ m.content.node = j
       /\ ~Covers(LeafSets[i], j)
       /\ MessagePool' = (MessagePool \ {m}) \cup 
            \* On joins, routing tables are updated with information from
            \* each hop. This is not implemented here.
            IF nh # i THEN {msg1} ELSE {msg2}
  /\ UNCHANGED <<Status, RoutingTables, LeafSets, Probing, Failed, ToJoin, Responsible>>

ReceiveJoinRequest(i) ==  
(* Node i receives a join request addressed to it. *)
  /\ Status[i] = "Ready" 
  /\ ToJoin[i] = i
  /\ \E m \in MessagePool:
       /\ m.content.type = "JoinRequest"
       /\ m.destination = i
       /\ Covers(LeafSets[i], m.content.node)
       /\ LET cont == [type |-> "JoinReply",  
                       rt |-> m.content.rt, 
                       ls |-> LeafSets[i]]
              msg == [destination |-> m.content.node, content |-> cont]
          IN 
          MessagePool' = (MessagePool \ {m}) \cup {msg}
       /\ ToJoin' = [ToJoin EXCEPT ![i] = m.content.node]
       /\ LeafSets'= [LeafSets EXCEPT ![i] = AddToLS({m.content.node}, @)]
  /\ UNCHANGED <<Status, RoutingTables, Probing, Failed, Responsible>>                 
                 
ReceiveJoinReply(i) == 
(* Node i receives a join reply. It updates its routing table and leaf set
   with the received information and probes the new nodes in its leaf set. *)
  /\ Status[i] = "Waiting" 
  /\ \E m \in MessagePool: 
       /\ m.content.type = "JoinReply"
       /\ m.destination = i
       /\ LET newRT == AddToRTWithSelf(LeafSetContent(m.content.ls)
                             \cup RTContent(m.content.rt), RoutingTables[i], i)
              newLS == AddToLS(LeafSetContent(m.content.ls), LeafSets[i])
              toProbe  == LeafSetContent(newLS) \ {i}
              probeMessages == ProbeSet(i, newLS, {}, toProbe)
          \* i needs to send out probe messages to change its state to "Ready".
          IN
          /\ RoutingTables' = [RoutingTables  EXCEPT ![i] = newRT] 
          /\ LeafSets'   = [LeafSets EXCEPT ![i] = newLS] 
          /\ Probing'= [Probing EXCEPT ![i] = toProbe]
          /\ MessagePool'= (MessagePool \ {m}) \cup probeMessages
          /\ Responsible' = [Responsible EXCEPT ![i] = m.content.ls.node]
  /\ UNCHANGED <<Status, ToJoin, Failed>>

-------------------------------------------------------------------------------

(* Lookup *)

Lookup(i, j) ==
(* i sends a request to find (the closest live node to) j. *)
  LET msg == [destination |-> i, 
              content |-> [type |-> "Lookup", node |-> j]] 
  IN
  /\ Status[i] # "Dead"
  /\ i # j
  /\ MessagePool' = MessagePool \cup {msg}
  /\ UNCHANGED <<Status, RoutingTables, LeafSets, Probing, Failed, ToJoin, Responsible>>

RouteLookup(i, j) == 
  /\ Status[i] # "Dead"
  /\ \E m \in MessagePool:
       /\ m.content.type ="Lookup" 
       /\ m.destination = i
       /\ m.content.node = j 
       /\ ~Covers(LeafSets[i], j)
       /\ LET nh == FindNext(i, j)
              msg1 == [destination |-> nh, content |-> m.content]
              msg2 == [destination |-> i,
                       content |-> [type |-> "NoLegalRoute", key |-> j]]
          IN 
          MessagePool' = (MessagePool \{m}) \cup 
            IF nh # i THEN {msg1} ELSE {msg2}
  /\ UNCHANGED <<Status, RoutingTables, LeafSets, Probing, Failed, ToJoin, Responsible>>

DeliverLookup(i, j) == 
(* Node i delivers the lookup message to j. A node may only deliver a lookup
   message whose destination is leased by this node. *)
  /\ Status[i] = "Ready" 
  /\ \E m \in MessagePool: 
       /\ m.content.type = "Lookup"
       /\ m.destination = i
       /\ m.content.node = j
       /\ Covers(LeafSets[i], j)
       /\ MessagePool'= (MessagePool \{m}) 
  /\ UNCHANGED <<Status, RoutingTables, LeafSets, Probing, Failed, ToJoin, Responsible>>
                  
-------------------------------------------------------------------------------

(* Probe *)

ReceiveProbeInner(i, m) ==
  LET j == m.content.node
      newFailed == Failed[i] \ {j} 
      newLS == AddToLS({j}, LeafSets[i])
      potentialMembers == LeafSetContent(AddToLS((LeafSetContent(m.content.ls) \ newFailed), newLS))
      (*prb1 == LeafSetContent(newLS) \cap m.content.failed*)
      \* NA: Excluded i because node does not need to probe itself:
      prb1 == (LeafSetContent(newLS) \cap m.content.failed) \ {i}
        \* Confirm if reportedly failed nodes have actually failed before
        \* removing them from the leaf set.
      prb2 == potentialMembers \ LeafSetContent(newLS)
        \* Confirm that new candidates for the leaf set are actually alive before
        \* adding them to the leaf set.
      toProbe == (prb1 \cup prb2) \ (Probing[i] \cup newFailed \cup {j})
        \* NA: No need to probe j, since it just probed.
      newmc == [type |-> "ProbeReply",  
                node |-> i,
                ls |-> LeafSets[i],
                failed |-> newFailed]
  IN
  /\ Failed' = [Failed EXCEPT ![i] = newFailed]
  /\ RoutingTables' = [RoutingTables EXCEPT ![i] = AddToRTWithSelf({j}, @, i)]
  /\ LeafSets' = [LeafSets EXCEPT ![i] = newLS]
  /\ Probing' = [Probing EXCEPT ![i] = @ \cup toProbe]
  /\ MessagePool' = (MessagePool  
       \cup { [destination |-> j, content |-> newmc] } 
       \cup ProbeSet(i, newLS, newFailed, toProbe)) \ {m}
  /\ UNCHANGED <<Status, ToJoin, Responsible>>

ReceiveProbe(i) ==
(* Node i receives a probe message from j. It updates its local information
   w.r.t. j and replies to j's message.*)
  /\ Status[i] # "Dead"
  /\ Status[i] = "Ready" \/ LeafSets[i] # EmptyLS(i)
      \* Prevents a node from responding to probes before receiving its
          \* Join Reply message.
  /\ \E m \in MessagePool:
       /\ m.content.type = "Probe"
       /\ m.destination = i
       /\ ReceiveProbeInner(i, m)

ReceiveProbeReplyInner(i,m) ==
  LET j == m.content.node
      ls == m.content.ls
      newFailed == Failed[i] \ {j} 
      newLS == AddToLS({j}, LeafSets[i]) 
      potentialMembers == LeafSetContent(AddToLS((LeafSetContent(ls) \ newFailed), newLS))
      \* NA: Excluded i because node does not need to probe itself.
      prb1 == (LeafSetContent(newLS) \cap m.content.failed ) \ (Probing[i] \cup newFailed \cup {i})
      prb2 == potentialMembers \ (LeafSetContent(newLS) \cup Probing[i] \cup newFailed \cup prb1)
      toProbe == (Probing[i] \cup prb1 \cup prb2) \ {j} 
      FinishedProbing ==
        /\ Status[i] = "Waiting" 
        /\ toProbe = {}
       \* /\ Overlaps(newLS) \/ IsComplete(newLS): Only useful when node failure is allowed (?)
      msg == [destination |-> Responsible[i], 
               content |-> [type |-> "LeaseRequest", node |-> i]]
  IN 
  /\ RoutingTables' =
       [RoutingTables  EXCEPT ![i] = AddToRTWithSelf({j}, @, i)]
  /\ LeafSets'   = [LeafSets    EXCEPT ![i] = newLS]
  /\ Failed' = [Failed  EXCEPT ![i] = IF FinishedProbing THEN {} ELSE newFailed]       
  /\ Probing'= [Probing EXCEPT ![i] = toProbe]
  /\ Status' = [Status  EXCEPT ![i] = IF FinishedProbing THEN "Ready" ELSE @]
  /\ MessagePool' = (MessagePool 
       \cup ProbeSet(i, newLS, newFailed, prb1)
       \cup ProbeSet(i, newLS, newFailed, prb2)
       \cup IF FinishedProbing THEN {msg} ELSE {}) \ {m} 
  /\ UNCHANGED <<ToJoin, Responsible>>

ReceiveProbeReply(i) ==
  /\ Status[i] # "Dead"
  /\ \E m \in MessagePool:
       /\ m.content.type = "ProbeReply"  
       /\ m.destination = i
       /\ ReceiveProbeReplyInner(i,m)

ReceiveLeaseRequest(i) ==
  /\ Status[i] = "Ready"
  /\ \E m \in MessagePool:
        LET msg == [destination |-> m.content.node, 
                    content |-> [type |-> "LeaseReply",
                                 ls |-> LeafSets[i],
                                 grant|-> TRUE]]
        IN
        /\ m.content.type = "LeaseRequest" 
        /\ m.destination = i
        /\ m.content.node = ToJoin[i]
        /\ ToJoin' = [ToJoin EXCEPT ![i] = i]
        /\ MessagePool' = (MessagePool \ {m}) \cup {msg}
  /\ UNCHANGED <<Status, RoutingTables, LeafSets, Probing, Failed,
                Responsible>> 

ReceiveLeaseReply(i) ==
  /\ Status[i] = "Ready"
  /\ \E m \in MessagePool:
        /\ m.content.type = "LeaseReply" 
        /\ m.destination = i
        /\ m.content.ls.node = Responsible[i]
        /\ Responsible' = [Responsible EXCEPT ![i] = i]
        /\ MessagePool' = (MessagePool \ {m})
  /\ UNCHANGED <<Status, RoutingTables, LeafSets, Probing, Failed,
                ToJoin>> 
===============================================================================
